/* unitopbm: convert a UTF-8 encoded Unicode file to a Portable Bit Map. */

/* Uses the GNU Unifont, cf. http://czyborra.com/unifont/ */

/* Environment variable UNIFONT must point to the file unifont.hex */

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <stdint.h> /* Replace by <sys/types.h> if no stdint.h */

#define DEFAULT_FONT_FILE_NAME "unifont.hex"
#define BUFSIZE 136

/* How to read UTF */
jmp_buf eof_exn;
jmp_buf invalid_utf_exn;

int32_t
fgetutf (FILE *f)
{
  int ch;
  int32_t val;

  ch = getc (f);
  if ( ch == EOF )
    longjmp (eof_exn, 1);
  if ( ch >= 0 && ch < 0x80 )
    return ch;
  if ( ch >= 0xc0 && ch < 0xe0 )
    {
      val = ((int32_t)(ch-0xc0))<<6;
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= (int32_t)(ch-0x80);
      else
	longjmp (invalid_utf_exn, 1);
      return val;
    }
  if ( ch >= 0xe0 && ch < 0xf0 )
    {
      val = ((int32_t)(ch-0xe0))<<12;
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<6;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= (int32_t)(ch-0x80);
      else
	longjmp (invalid_utf_exn, 1);
      return val;
    }
  if ( ch >= 0xf0 && ch < 0xf8 )
    {
      val = ((int32_t)(ch-0xf0))<<18;
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<12;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<6;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= (int32_t)(ch-0x80);
      else
	longjmp (invalid_utf_exn, 1);
      return val;
    }
  if ( ch >= 0xf8 && ch < 0xfc )
    {
      val = ((int32_t)(ch-0xf8))<<24;
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<18;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<12;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<6;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= (int32_t)(ch-0x80);
      else
	longjmp (invalid_utf_exn, 1);
      return val;
    }
  if ( ch >= 0xfc && ch < 0xfe )
    {
      val = ((int32_t)(ch-0xf8))<<30;
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<24;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<18;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<12;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= ((int32_t)(ch-0x80))<<6;
      else
	longjmp (invalid_utf_exn, 1);
      ch = getc (f);
      if ( ch >= 0x80 && ch < 0xc0 )
	val |= (int32_t)(ch-0x80);
      else
	longjmp (invalid_utf_exn, 1);
      return val;
    }
  longjmp (invalid_utf_exn, 1);
}

/* The ``stuff'' library */
typedef struct stuff_s {
  size_t cell_size;
  size_t data_size;
  char *data;
} stuff_t;

void
init_stuff (stuff_t *stuff, size_t cell_size)
{
  stuff->cell_size = cell_size;
  stuff->data_size = 0;
  stuff->data = NULL;
}

void
stuff_put (stuff_t *stuff, size_t index, const void *cell)
{
  if ( (index+1)*stuff->cell_size > stuff->data_size )
    {
      size_t new_size;

      new_size = (index+1024)*stuff->cell_size;
      stuff->data = realloc (stuff->data, new_size);
      if ( ! stuff->data )
	{
	  fprintf (stderr, "Out of memory!\n");
	  exit (1);
	}
      memset (stuff->data + stuff->data_size, 0,
	      new_size - stuff->data_size);
      stuff->data_size = new_size;
    }
  memcpy (stuff->data + index*stuff->cell_size, cell, stuff->cell_size);
}

void
stuff_get (stuff_t *stuff, size_t index, void *cell)
{
  if ( index*stuff->cell_size < stuff->data_size )
    memcpy (cell, stuff->data + index*stuff->cell_size, stuff->cell_size);
  else
    memset (cell, 0, stuff->cell_size);
}

/* Image */
stuff_t image;
int size_x;
int size_y;

typedef struct {
  char defn;
  stuff_t line;
} image_line_t;

void
init_image (void)
{
  init_stuff (&image, sizeof (image_line_t));
  size_x = 0;
  size_y = 0;
}

void
image_put (int x, int y, char b)
{
  image_line_t l;

  stuff_get (&image, y, &l);
  if ( ! l.defn )
    {
      l.defn = 1;
      init_stuff (&l.line, sizeof (char));
    }
  stuff_put (&l.line, x, &b);
  stuff_put (&image, y, &l);
  if ( x >= size_x )
    size_x = x + 1;
  if ( y >= size_y )
    size_y = y + 1;
}

char
image_get (int x, int y)
{
  image_line_t l;
  char b;

  stuff_get (&image, y, &l);
  if ( ! l.defn )
    return 0;
  stuff_get (&l.line, x, &b);
  return b;
}

void
image_dump (void)
{
  int x, y;
  printf ("P1\n%d %d\n", size_x, size_y);
  for ( y=0 ; y<size_y ; y++ )
    {
      for ( x=0 ; x<size_x ; x++ )
	printf ("%d ", image_get (x, y));
      printf ("\n");
    }
}

/* Main part */
char *font_file_name;
FILE *font_file;
stuff_t font_lines;

typedef struct {
  char defn;
  fpos_t p;
} e_fpos_t;

FILE *inf;

void
perror_abort (char *msg)
{
  perror (msg);
  exit (1);
}

int
hex_digit (char c)
{
  if ( c >= '0' && c <= '9' )
    return c-'0';
  if ( c >= 'A' && c <= 'f' )
    return c-'A'+10;
  if ( c >= 'a' && c <= 'f' )
    return c-'a'+10;
  fprintf (stderr, "Non hexadecimal digit %c\n", c);
  exit (1);
}

int
main (int argc, char *argv[])
{
  int x, y;

  /* Font file */
#if 0
  /* This won't work.  BLOODY SHIT THIS LANGUAGE IS STUPID. */
  font_file_name = getenv ("UNIFONT") || DEFAULT_FONT_FILE_NAME;
#else
  font_file_name = getenv ("UNIFONT");
  if ( ! font_file_name )
    font_file_name = DEFAULT_FONT_FILE_NAME;
#endif

  font_file = fopen (font_file_name, "r");
  if ( ! font_file )
    perror_abort (font_file_name);
  init_stuff (&font_lines, sizeof (e_fpos_t));
  while ( ! feof (font_file) )
    {
      char s[BUFSIZE];
      e_fpos_t pos;
      int index;

      pos.defn = 1;
      fgetpos (font_file, &pos.p);
      fgets (s, BUFSIZE, font_file);
      if ( s[0] == '#' )
	continue;
      index = (hex_digit (s[0]) << 12)
	| (hex_digit (s[1]) << 8)
	| (hex_digit (s[2]) << 4)
	| hex_digit (s[3]);
      stuff_put (&font_lines, index, &pos);
    }
  /* Input file */
  if ( argc > 1 )
    {
      inf = fopen (argv[1], "r");
      if ( ! inf )
	perror_abort (argv[1]);
    }
  else
    inf = stdin;
  if ( setjmp (eof_exn) )
    {
      image_dump ();
      exit (0);
    }
  if ( setjmp (invalid_utf_exn) )
    {
      fprintf (stderr, "Error: invalid UTF data.\n");
      exit (1);
    }
  init_image ();
  x = 0;
  y = 0;
  while (1) /* Exit from this loop by calling an exception */
    {
      int32_t ch;
      e_fpos_t pos;
      char s[BUFSIZE];
      int i;

      ch = fgetutf (inf);
      if ( ch == '\n' )
	{
	  x = 0;
	  y += 16;
	  continue;
	}
      stuff_get (&font_lines, ch, &pos);
      if ( ! pos.defn )
	{
	  fprintf (stderr, "Warning: character %ld not defined.\n",
		   (long)ch);
	  continue;
	}
      fsetpos (font_file, &pos.p);
      fgets (s, BUFSIZE, font_file);
      if ( strlen (s) == 38 )
	{
	  for ( i=0 ; i<16 ; i++ )
	    {
	      image_put (x+0, y+i, (hex_digit (s[i*2+5])>>3) & 1);
	      image_put (x+1, y+i, (hex_digit (s[i*2+5])>>2) & 1);
	      image_put (x+2, y+i, (hex_digit (s[i*2+5])>>1) & 1);
	      image_put (x+3, y+i, hex_digit (s[i*2+5]) & 1);
	      image_put (x+4, y+i, (hex_digit (s[i*2+6])>>3) & 1);
	      image_put (x+5, y+i, (hex_digit (s[i*2+6])>>2) & 1);
	      image_put (x+6, y+i, (hex_digit (s[i*2+6])>>1) & 1);
	      image_put (x+7, y+i, hex_digit (s[i*2+6]) & 1);
	    }
	  x+=8;
	}
      else if ( strlen (s) == 70 )
	{
	  for ( i=0 ; i<16 ; i++ )
	    {
	      image_put (x+0, y+i, (hex_digit (s[i*4+5])>>3) & 1);
	      image_put (x+1, y+i, (hex_digit (s[i*4+5])>>2) & 1);
	      image_put (x+2, y+i, (hex_digit (s[i*4+5])>>1) & 1);
	      image_put (x+3, y+i, hex_digit (s[i*4+5]) & 1);
	      image_put (x+4, y+i, (hex_digit (s[i*4+6])>>3) & 1);
	      image_put (x+5, y+i, (hex_digit (s[i*4+6])>>2) & 1);
	      image_put (x+6, y+i, (hex_digit (s[i*4+6])>>1) & 1);
	      image_put (x+7, y+i, hex_digit (s[i*4+6]) & 1);
	      image_put (x+8, y+i, (hex_digit (s[i*4+7])>>3) & 1);
	      image_put (x+9, y+i, (hex_digit (s[i*4+7])>>2) & 1);
	      image_put (x+10, y+i, (hex_digit (s[i*4+7])>>1) & 1);
	      image_put (x+11, y+i, hex_digit (s[i*4+7]) & 1);
	      image_put (x+12, y+i, (hex_digit (s[i*4+8])>>3) & 1);
	      image_put (x+13, y+i, (hex_digit (s[i*4+8])>>2) & 1);
	      image_put (x+14, y+i, (hex_digit (s[i*4+8])>>1) & 1);
	      image_put (x+15, y+i, hex_digit (s[i*4+8]) & 1);
	    }
	  x+=16;
	}
      else
	{
	  fprintf (stderr, "Warning: wrong font line length %d\n",
		   strlen (s));
	  continue;
	}
    }
}
